package com.fr.plugins.dsgroup;

import com.fr.base.BaseUtils;

import com.fr.base.TableData;
import com.fr.config.utils.UniqueKey;
import com.fr.design.actions.server.GlobalTableDataAction;

import com.fr.design.data.DesignTableDataManager;
import com.fr.design.data.datapane.TableDataSourceOP;
import com.fr.design.data.datapane.TableDataTree;

import com.fr.design.data.datapane.TableDataTreePane;
import com.fr.design.data.tabledata.wrapper.TemplateTableDataWrapper;
import com.fr.design.gui.itree.refreshabletree.UserObjectOP;
import com.fr.file.DatasourceManager;
import com.fr.file.DatasourceManagerProvider;
import com.fr.file.ProcedureConfig;
import com.fr.file.TableDataConfig;
import com.fr.general.NameObject;
import com.fr.design.constants.UIConstants;
import com.fr.design.data.tabledata.wrapper.TableDataWrapper;
import com.fr.design.gui.ilable.UILabel;
import com.fr.design.gui.itree.refreshabletree.ExpandMutableTreeNode;
import com.fr.design.icon.IconPathConstants;
import com.fr.general.ComparatorUtils;
import com.fr.log.FineLoggerFactory;
import com.fr.plugins.dsgroup.fun.DSGroupTableData;
import com.fr.third.apache.log4j.Logger;

import javax.swing.*;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;
import java.awt.*;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Created by phetran_ye on 2018/8/12.
 */
public class CustomTableDataTree extends TableDataTree {
    private static final long serialVersionUID = 1L;

    private CustomTableDataTreePane pane;
    public void setPane(CustomTableDataTreePane p){
        this.pane=p;
    }
    /**
     * Constructor.
     */

    // CellRenderer
    private DefaultTreeCellRenderer tableDataTreeCellRenderer = new DefaultTreeCellRenderer() {
        private static final long serialVersionUID = 1L;

        @Override
        public Component getTreeCellRendererComponent(JTree tree, Object value, boolean selected, boolean expanded, boolean leaf, int row, boolean hasFocus) {
            super.getTreeCellRendererComponent(tree, value, selected, expanded, leaf, row, hasFocus);
            ExpandMutableTreeNode treeNode = (ExpandMutableTreeNode) value;
            Object userObj = treeNode.getUserObject();
            if (userObj instanceof String) {
                // p:这个是column field.
                this.setIcon(BaseUtils.readIcon("com/fr/design/images/data/field.png"));
                this.setText((String) userObj);
            } else if (userObj instanceof NameObject) {
                NameObject nameObject = (NameObject) userObj;
                this.setText(nameObject.getName());
                if (nameObject.getObject() instanceof TableDataWrapper) {
                    TableDataWrapper tableDataWrappe = (TableDataWrapper) nameObject.getObject();
                    this.setIcon(tableDataWrappe.getIcon());
                } else if (nameObject.getObject() instanceof Integer) {
                    int num = (Integer) nameObject.getObject();
                    if (num == TableDataSourceOP.SERVER_TABLE_DATA) {
                        this.setIcon(BaseUtils.readIcon(IconPathConstants.STD_SHOW_ICON_PATH));
                    } else if (num == TableDataSourceOP.STORE_PRECEDURE_DATA) {
                        this.setIcon(BaseUtils.readIcon(IconPathConstants.SP_SHOW_ICON_PATH));
                    } else {
                        this.setIcon(BaseUtils.readIcon(IconPathConstants.DS_QUERY_ICON_PATH));
                    }
                } else {
                    this.setIcon(BaseUtils.readIcon("/com/fr/design/images/data/store_procedure.png"));
                }
            } else if (userObj == PENDING) {
                this.setIcon(null);
                this.setText(PENDING.toString());
            }
            // 这里新建一个Label作为render是因为JTree在动态刷新的时候，节点上render画布的的宽度不会变，会使得一部分比较长的数据显示为"..."
            UILabel label = new UILabel();
            label.setText(getText());
            label.setIcon(getIcon());
            Dimension dim = label.getPreferredSize();
            dim.height += 2;
            this.setSize(dim);
            this.setPreferredSize(dim);
            this.setBackgroundNonSelectionColor(UIConstants.TREE_BACKGROUND);
            this.setTextSelectionColor(Color.WHITE);
            this.setBackgroundSelectionColor(UIConstants.FLESH_BLUE);
            return this;
        }
    };

    public DefaultTreeCellRenderer getTableDataTreeCellRenderer() {
        return tableDataTreeCellRenderer;
    }

    public void setTableDataTreeCellRenderer(DefaultTreeCellRenderer tableDataTreeCellRenderer) {
        this.tableDataTreeCellRenderer = tableDataTreeCellRenderer;
    }

    protected void refreshTreeNode(ExpandMutableTreeNode eTreeNode, String childName) {
        if (interceptRefresh(eTreeNode)) {
            return;
        }
        boolean refreshall = childName.isEmpty();
        ExpandMutableTreeNode[] new_nodes = loadChildTreeNodes(eTreeNode);

        java.util.List<DefaultMutableTreeNode> childTreeNodeList = new java.util.ArrayList<DefaultMutableTreeNode>();
        for (int i = 0, len = eTreeNode.getChildCount(); i < len; i++) {
            if (eTreeNode.getChildAt(i) instanceof ExpandMutableTreeNode) {
                childTreeNodeList.add((ExpandMutableTreeNode) eTreeNode.getChildAt(i));
            } else {
                childTreeNodeList.add((DefaultMutableTreeNode) eTreeNode.getChildAt(i));
            }
        }

        eTreeNode.removeAllChildren();

        for (int ci = 0; ci < new_nodes.length; ci++) {
            Object cUserObject = new_nodes[ci].getUserObject();
            ExpandMutableTreeNode cTreeNode = null;
            for (int ni = 0, nlen = childTreeNodeList.size(); ni < nlen; ni++) {
                cTreeNode = (ExpandMutableTreeNode) childTreeNodeList.get(ni);
                if (ComparatorUtils.equals(cTreeNode.getUserObject(), cUserObject)) {
                    if (!refreshall && !ComparatorUtils.equals(childName, ((NameObject) cUserObject).getName())) {
                        new_nodes[ci] = cTreeNode;
                        break;
                    }
                    new_nodes[ci].setExpanded(cTreeNode.isExpanded());
                    if (cTreeNode.getChildCount()>0 && cTreeNode.getFirstChild() instanceof ExpandMutableTreeNode && cTreeNode.isExpanded()) {
                        checkChildNodes(cTreeNode, new_nodes[ci]);
                    }
                    break;
                }
            }

            //eTreeNode.add(new_nodes[ci]);
        }
        loadTreeNode(eTreeNode);
    }


    protected void checkChildNodes(ExpandMutableTreeNode oldNode, ExpandMutableTreeNode newNode) {
        for (int i = 0; i < oldNode.getChildCount(); i++) {
            ExpandMutableTreeNode oldChild = (ExpandMutableTreeNode) oldNode.getChildAt(i);
            for (int j = 0; j < newNode.getChildCount(); j++) {
                ExpandMutableTreeNode newChild = (ExpandMutableTreeNode) newNode.getChildAt(j);
                newChild.removeAllChildren();
                ExpandMutableTreeNode[] nodes = CustomTableDataTree.this.loadChildTreeNodes(newChild);
                for (int k = 0; k < nodes.length; k++) {
                    newChild.add(nodes[k]);
                }
                if (newChild.getChildCount() > 1 && ((ExpandMutableTreeNode) newChild.getFirstChild()).getUserObject() == PENDING) {
                    newChild.remove(0);
                }
                if (ComparatorUtils.equals(oldChild.getUserObject(), newChild.getUserObject())) {
                    newChild.setExpanded(oldChild.isExpanded());
                }
            }
        }
    }

    /*
     * p:获得选中的NameObject = name + tabledata.
	 */
    public NameObject getSelectedNameObject() {
        TreePath selectedTreePath = this.getSelectionPath();
        if (selectedTreePath == null) {
            return null;
        }
        ExpandMutableTreeNode selectedTreeNode = (ExpandMutableTreeNode) selectedTreePath.getLastPathComponent();
        Object selectedUserObject = selectedTreeNode.getUserObject();
        ExpandMutableTreeNode parentTreeNode = (ExpandMutableTreeNode) selectedTreeNode.getParent();
        Object parentUserObject = parentTreeNode.getUserObject();
        if (parentUserObject instanceof NameObject && ((NameObject) parentUserObject).getObject() instanceof Integer) {
            if (selectedUserObject instanceof NameObject) {
                return (NameObject) selectedUserObject;
            }
        } else {
            parentTreeNode = (ExpandMutableTreeNode) selectedTreeNode.getParent();
            parentUserObject = parentTreeNode.getUserObject();

            if (parentUserObject != null) {
                if (!(parentUserObject instanceof NameObject))  {
                    return (NameObject) selectedUserObject;
                }
                //分组数据集下的子节点返回自身
                else if(isGroupData(selectedTreeNode)){
                    return (NameObject) selectedUserObject;
                }
                else {
                    return (NameObject) parentUserObject;
                }
            }
        }
        return null;

    }

    public TableDataWrapper[] getSelectedDatas() {
        TreePath[] selectedTreePaths = this.getSelectionPaths();
        if (selectedTreePaths == null || selectedTreePaths.length == 0) {
            return null;
        }
        TableDataWrapper[] nameobjs = new TableDataWrapper[selectedTreePaths.length];
        for (int i = 0; i < selectedTreePaths.length; i++) {
            TreePath selectedTreePath = selectedTreePaths[i];
            ExpandMutableTreeNode selectedTreeNode = (ExpandMutableTreeNode) selectedTreePath.getLastPathComponent();
            Object selectedUserObject = selectedTreeNode.getUserObject();
            ExpandMutableTreeNode parentTreeNode = (ExpandMutableTreeNode) selectedTreeNode.getParent();
            Object parentUserObject = parentTreeNode.getUserObject();
            if (parentUserObject instanceof NameObject && ((NameObject) parentUserObject).getObject() instanceof Integer) {
                if (selectedUserObject instanceof NameObject) {
                    Object obj = ((NameObject) selectedUserObject).getObject();
                    if (obj instanceof TableDataWrapper) {
                        nameobjs[i] = (TableDataWrapper) obj;
                    }
                }
            } else {
                return new TableDataWrapper[0];
            }
        }

        return nameobjs;
    }

    public NameObject getRealSelectedNameObject() {
        TreePath selectedTreePath = this.getSelectionPath();
        if (selectedTreePath == null) {
            return null;
        }

        ExpandMutableTreeNode selectedTreeNode = (ExpandMutableTreeNode) selectedTreePath.getLastPathComponent();
        Object selectedUserObject = selectedTreeNode.getUserObject();
        if (selectedUserObject instanceof NameObject) {
            return (NameObject) selectedUserObject;
        }

        selectedTreeNode = (ExpandMutableTreeNode) selectedTreeNode.getParent();
        selectedUserObject = selectedTreeNode.getUserObject();
        if (selectedUserObject instanceof NameObject) {
            return (NameObject) selectedUserObject;
        }
        return null;
    }

    /**
     * p:添加一个NameObject节点.
     */
    public void addNameObject(NameObject no) {
        if (no == null) {
            return;
        }
        DefaultTreeModel treeModel = (DefaultTreeModel) this.getModel();

        // 新建一个放着NameObject的newChildTreeNode,加到Root下面
        ExpandMutableTreeNode root = (ExpandMutableTreeNode) treeModel.getRoot();

        ExpandMutableTreeNode newChildTreeNode = new ExpandMutableTreeNode(no);
        root.add(newChildTreeNode);
        newChildTreeNode.add(new ExpandMutableTreeNode());

        treeModel.reload(root);
    }

    public <T extends UserObjectOP<?>> void  populate(T userObject,Object o) {
        DefaultTreeModel treeModel = (DefaultTreeModel) this.getModel();
        ExpandMutableTreeNode root = (ExpandMutableTreeNode) treeModel.getRoot();
        root.setUserObject(userObject);
        root.removeAllChildren();

        loadTreeNode(root);

        treeModel.reload(root);
        root.expandCurrentTreeNode(this);
    }

    public String getDSName(ExpandMutableTreeNode tn){
        return ((NameObject)tn.getUserObject()).getName();
    }

    public DSGroupTableData getGroupData(ExpandMutableTreeNode tn){
        return (DSGroupTableData)((TableDataWrapper) ((NameObject)tn.getUserObject()).getObject()).getTableData();
    }

    public boolean isGroupData(ExpandMutableTreeNode tn){
        if(null==tn) return false;
        if(!(tn.getUserObject() instanceof NameObject)) return false;
        return ((TableDataWrapper) ((NameObject)tn.getUserObject()).getObject()).getTableData() instanceof DSGroupTableData;
    }

    private void loadTreeNode(ExpandMutableTreeNode root){
        Map<String,ExpandMutableTreeNode> treeNodeMap=new HashMap<String,ExpandMutableTreeNode>();
        Map<String,ExpandMutableTreeNode> preloadTreeNodeMap=new HashMap<String,ExpandMutableTreeNode>();
        ExpandMutableTreeNode[] children = loadChildTreeNodes(root);
        for(ExpandMutableTreeNode etn:children){
            treeNodeMap.put(getDSName(etn),etn);
            preloadTreeNodeMap.put(getDSName(etn),etn);
        }
        for(ExpandMutableTreeNode etn:children){
            if(isGroupData(etn)) loadDSGroupTreeNode(etn, treeNodeMap,preloadTreeNodeMap);
        }

        for(Map.Entry<String,ExpandMutableTreeNode> en:preloadTreeNodeMap.entrySet()){
            root.add(en.getValue());
        }
    }

    private void loadDSGroupTreeNode(ExpandMutableTreeNode groupNode,Map<String,ExpandMutableTreeNode> nodes,Map<String,ExpandMutableTreeNode> leftNodes){
        if(getGroupData(groupNode).getChildrenDSNames().equals("")) return;
        String[] nodeNames=getGroupData(groupNode).getChildrenDSNames().split(DSGroupTableData.getSeparator());
        groupNode.removeAllChildren();
        if(!getGroupData(groupNode).getParentDSName().equals("")) leftNodes.remove(groupNode);
        for(int i=0;i<nodeNames.length;i++){
            if(isGroupData(nodes.get(nodeNames[i])))
                loadDSGroupTreeNode(nodes.get(nodeNames[i]), nodes, leftNodes);
            groupNode.add(nodes.get(nodeNames[i]));
            leftNodes.remove(nodeNames[i]);
        }
    }

    public void moveToGroup(ExpandMutableTreeNode groupToMoveTo){

        //test
        /*DefaultTreeModel treeModel = (DefaultTreeModel) this.getModel();
        ExpandMutableTreeNode root = (ExpandMutableTreeNode) treeModel.getRoot();
        root.setUserObject(pane.getOp());
        ExpandMutableTreeNode[] nodes = loadChildTreeNodes(root);
        //DesignTableDataManager.clearGlobalDs();
        for (int i = 0; i < nodes.length; i++) {
            NameObject nameObject = (NameObject) nodes[i].getUserObject();
            TableData tableData=((TableDataWrapper) nameObject.getObject()).getTableData();
            //tableDataConfig.addTableData(nameObject.getName(), ((TableDataWrapper) nameObject.getObject()).getTableData());
        }

        TableDataConfig tableDataConfig=TableDataConfig.getInstance();
        TableData tableData=tableDataConfig.getTableData("数据库查询1");
        Map<String,TableData> tableDatasd=tableDataConfig.getTableDatas();
        tableDataConfig.removeAllTableData();
        Map<String,TableData> tableDataMap = new LinkedHashMap<String,TableData>();

        tableDataMap.put(((NameObject) groupToMoveTo.getUserObject()).getName(),
                (TableData) ((TableDataWrapper) ((NameObject) groupToMoveTo.getUserObject()).getObject()).getTableData()
        );

        tableDataConfig.setTableDatas(tableDataMap);
        Map<String,TableData> tableDatas=tableDataConfig.getTableDatas();*/
        //end



        String children=getGroupData(groupToMoveTo).getChildrenDSNames();
        TreePath[] paths=getSelectionPaths();
        String chosenChildren="";
        //将选择的节点拼成子节点数据,以防止选择了分组又选择了其下的数据集
        for(int i=0;i<paths.length;i++){
            //分组本身
            if(((ExpandMutableTreeNode) paths[i].getLastPathComponent()).equals(groupToMoveTo)) continue;
            //待分组的节点是分组
            if(isGroupData((ExpandMutableTreeNode) paths[i].getLastPathComponent())){
                chosenChildren+=DSGroupTableData.getSeparator()+getGroupData((ExpandMutableTreeNode) paths[i].getLastPathComponent()).getChildrenDSNames();
            }
        }
        chosenChildren+=DSGroupTableData.getSeparator();
        for(int i=0;i<paths.length;i++){
            //分组本身
            if(((ExpandMutableTreeNode) paths[i].getLastPathComponent()).equals(groupToMoveTo)) continue;
            //待分组的节点是分组
            if(isGroupData((ExpandMutableTreeNode) paths[i].getLastPathComponent())){
                getGroupData((ExpandMutableTreeNode) paths[i].getLastPathComponent()).setParentDSName(
                        getDSName(groupToMoveTo)
                );

            }
            //选中的节点是该分组下的节点
            if(chosenChildren.indexOf(
                    DSGroupTableData.getSeparator()+
                            getDSName((ExpandMutableTreeNode) paths[i].getLastPathComponent())+
                            DSGroupTableData.getSeparator()
            )>-1
                    ) continue;
            if(i==0){
                children+=((children==null||children.equals(""))?"":DSGroupTableData.getSeparator())+
                        getDSName((ExpandMutableTreeNode) paths[i].getLastPathComponent());
            }else {
                children+= ((children==null||children.equals(""))?"":DSGroupTableData.getSeparator())+
                        getDSName((ExpandMutableTreeNode) paths[i].getLastPathComponent());
            }
            deleteNodeToMoveFromParent((ExpandMutableTreeNode) paths[i].getLastPathComponent());
        }
        getGroupData(groupToMoveTo).setChildrenDSNames(children);
        TableDataSourceOP op=pane.getOp();
        fireChangeEvent();

        refreshChildByName(getDSName(groupToMoveTo));
    }

    public void moveToRoot(){
        TreePath[] paths=getSelectionPaths();
        for(int i=0;i<paths.length;i++){
            ExpandMutableTreeNode selectedNode=(ExpandMutableTreeNode) paths[i].getLastPathComponent();
            //如果选中的节点有字段直接返回  防止拖动字段值影响拖动
            if(!(selectedNode.getUserObject() instanceof NameObject)) return;
            deleteNodeToMoveFromParent(selectedNode);
        }
        fireChangeEvent();
        refresh();
    }

    private void fireChangeEvent(){
        TableDataSourceOP op=pane.getOp();
        if (op!=null) {
            if(op.getDataMode()==0){
                pane.getTC().fireTargetModified();
            }else if(op.getDataMode()==1){//服务器数据集直接保存
                TableDataConfig tableDataConfig = TableDataConfig.getInstance();
                DefaultTreeModel treeModel = (DefaultTreeModel) this.getModel();
                ExpandMutableTreeNode root = (ExpandMutableTreeNode) treeModel.getRoot();
                root.setUserObject(pane.getOp());
                ExpandMutableTreeNode[] nodes = loadChildTreeNodes(root);

                LinkedHashMap<String,TableData> tableDatas = new LinkedHashMap();
                for (int i = 0; i < nodes.length; i++) {
                    if(isGroupData(nodes[i])){
                        NameObject nameObject = (NameObject) nodes[i].getUserObject();
                        TableData newData=((TableDataWrapper) nameObject.getObject()).getTableData();
                        ((UniqueKey)newData).setNameSpace(null);
                        //tableDataConfig.removeTableData(nameObject.getName());
                        ((DSGroupTableData)tableDataConfig.getTableData(nameObject.getName())).setAttrParentDSName(((DSGroupTableData)newData).getAttrParentDSName());
                        ((DSGroupTableData)tableDataConfig.getTableData(nameObject.getName())).setAttrChildrenDSNames(((DSGroupTableData)newData).getAttrChildrenDSNames());
                    }
                    /*NameObject nameObject = (NameObject) nodes[i].getUserObject();
                    TableData tableData=((TableDataWrapper) nameObject.getObject()).getTableData();
                    ((UniqueKey)tableData).setNameSpace(null);
                    tableDatas.put(nameObject.getName(),tableData);*/
                }
                //tableDataConfig.setTableDatas(tableDatas);
                //new GlobalTableDataAction().actionPerformed(null);
                //new GlobalTableDataAction().fireDSChanged();

            }
        }
    }

    //删除节点原分组子节点信息
    protected void deleteNodeToMoveFromParent(ExpandMutableTreeNode node){
        DefaultTreeModel treeModel = (DefaultTreeModel) this.getModel();
        ExpandMutableTreeNode root = (ExpandMutableTreeNode) treeModel.getRoot();
        ExpandMutableTreeNode[] children = loadChildTreeNodes(root);
        String separator=DSGroupTableData.getSeparator();
        for(int i=0;i<children.length;i++){
            if(isGroupData(children[i])){
                String childNodes=getGroupData(children[i]).getChildrenDSNames();
                childNodes=separator+childNodes+separator;
                String nodeName=separator+getDSName(node)+separator;
                if(childNodes.indexOf(nodeName)>-1){
                    childNodes=childNodes.replace(nodeName,separator).replaceAll("^"+separator+"{1}","").replaceAll(separator+"${1}","");
                    getGroupData(children[i]).setChildrenDSNames(childNodes);
                }
            }
        }
    }

    //更新了某个分组数据集名称
    public void updateNodeName(String oldName,String newName){
        DefaultTreeModel treeModel = (DefaultTreeModel) this.getModel();
        ExpandMutableTreeNode root = (ExpandMutableTreeNode) treeModel.getRoot();
        ExpandMutableTreeNode[] children = loadChildTreeNodes(root);
        String separator=DSGroupTableData.getSeparator();
        String newNameToReplaceWith=separator+newName+separator;
        String oldNameToMatchWith=separator+oldName+separator;
        for(int i=0;i<children.length;i++){
            if(isGroupData(children[i])){
                DSGroupTableData tableData=getGroupData(children[i]);
                String childNodes=tableData.getChildrenDSNames();
                childNodes=separator+childNodes+separator;
                if(childNodes.indexOf(oldNameToMatchWith)>-1){
                    childNodes=childNodes.replace(oldNameToMatchWith,newNameToReplaceWith).replaceAll("^"+separator+"{1}","").replaceAll(separator+"${1}","");
                    getGroupData(children[i]).setChildrenDSNames(childNodes);
                }
                String parentNode=tableData.getParentDSName();
                if(parentNode.equals(oldName)){
                    parentNode=newName;
                    tableData.setParentDSName(parentNode);
                }
            }
        }
    }

}
